#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>


/* 
   == N-element chain with driving force ==

   Command-line args are:

   * N      the number of masses
   * k      the spring constant (m = 1 by definition)
   * ampl   the amplitude of the driving force
   * om_d   the angular frequency of the driving force
   * its    the number of time steps to simulate
   * dt     the time step

*/

int main (int argc, char *argv[]) {

  int rank, size;
  int ntot, n, i,  it, maxits, left, right;
  double k, omd, *x, *v, *tp, ampl, dt, maxamp, tmp;
  FILE *f1, *fN;
  MPI_Status stat;

  MPI_Init (&argc, &argv);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
  MPI_Comm_size (MPI_COMM_WORLD, &size);

  if (argc != 7) {
    if (rank == 0) printf("Should be six arguments! Exiting...\n");
    MPI_Finalize ();
    exit(2);
  }

  ntot   = atoi(argv[1]);
  k      = atof(argv[2]);
  ampl   = atof(argv[3]);
  omd    = atof(argv[4]);
  maxits = atoi(argv[5]);
  dt     = atof(argv[6]);

  if (ntot % size != 0) {
    if (rank == 0) printf("Numprocs must divide npoints evenly!\n");
    MPI_Finalize ();
    exit (3);
  }

  /* Number of points to simulate in this subdomain */
  n = ntot/size;
  
  /* Allocate arrays */
  x    = (double *) calloc (n + 2, sizeof(double));
  v    = (double *) calloc (n + 2, sizeof(double));

  for (i=0; i<=n+1; i++)
    x[i] = 0.0;

  /* Initial (t=0) velocities set here: */
  for (i=0; i<n+2; i++)
    v[i] = 0.0;

  /* For half-step method.  This is v[-1/2] = v[0] - (1/2) a[0] dt;
     the first iteration below then produces v[+1/2] as needed. */
  for (i=1; i<=n; i++)
    v[i] += 0.5*k*(2.0*x[i] - x[i-1] - x[i+1])*dt;

  /* Determine neighbors (periodic BC possible here!) */
  left = rank - 1;
  if (rank < size - 1)
    right = rank + 1;
  else right = -1;

  if (rank == 0)
    f1 = fopen("mass1-mpi", "w");
  if (rank == size - 1)
    fN = fopen("massN-mpi", "w");

  /* For dentifying resonance frequencies: */
/*   maxamp = 0.0; */

  /* Main loop */
  for (it = 0; it < maxits; it++) {

    /* Write output */
    if (it % 5 == 0) {
      if (rank == 0) fprintf (f1, "%lf\n", x[1]);
      if (rank == size - 1) fprintf (fN, "%lf\n", x[n]);
    }

    /* Update driving mass */
    if (rank == 0)
      x[0] = ampl*sin(omd*(it+1)*dt);

    /* Pass endpoint values to neighbors.  This should work even if
       the standard send is synchronous, i.e. it does not return until
       the message is received. */

    /* Odds first send, then receive: */
    if (rank % 2 != 0) {
      if (left >= 0) {
	MPI_Send (&x[1],1,MPI_DOUBLE,left,0,MPI_COMM_WORLD);
	MPI_Recv (&x[0],1,MPI_DOUBLE,left,1,MPI_COMM_WORLD,&stat);
      }
      if (right >= 0) {
	MPI_Send (&x[n],1,MPI_DOUBLE,right,1,MPI_COMM_WORLD);
	MPI_Recv (&x[n+1],1,MPI_DOUBLE,right,0,MPI_COMM_WORLD,&stat);
      }
    }
    else {
      if (right >= 0) {
	MPI_Recv (&x[n+1],1,MPI_DOUBLE,right,0,MPI_COMM_WORLD,&stat);
	MPI_Send (&x[n],1,MPI_DOUBLE,right,1,MPI_COMM_WORLD);
      }
      if (left >= 0) {
	MPI_Recv (&x[0],1,MPI_DOUBLE,left,1,MPI_COMM_WORLD,&stat);
	MPI_Send (&x[1],1,MPI_DOUBLE,left,0,MPI_COMM_WORLD);
      }
    }

    /* Evolution by half step method */
    for (i=1; i<=n; i++)
      v[i] += - k*(2.0*x[i] - x[i-1] - x[i+1])*dt;

    for (i=1; i<=n; i++)
      x[i] += v[i]*dt;

/*     if ((tmp = fabs(x[1])) > maxamp)  */
/*       maxamp = tmp; */
  }

  if (rank == 0) fclose (f1);
  if (rank == size - 1) fclose (fN);

/*   printf("Max displacement of mass 1 = %lf\n", maxamp); */

  MPI_Finalize ();

  return 0;
}
