#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <mpi.h>

#define N 50
#define M 51
#define OUTSKIP 1000

/* 
   == N x M element mattress with driving force(s) ==

   MPI version: uses simple strip decomposition in first dimension.

   Command-line args are:

   * k      the spring constant (m = 1 by definition)
   * ampl   the amplitude of the driving force
   * om_d   the angular frequency of the driving force
   * its    the number of time steps to simulate
   * dt     the time step

*/

int main (int argc, char *argv[]) {

  int rank, size;
  int n, m, i, j, it, maxits, up, dn;
  double k, omd, **x, **v, ampl, dt, maxamp, tmp;
  char outbase[5] = "out.";
  char outfile[15];
  char outsuffix[10];
  FILE *fp;
  MPI_Status stat;

  MPI_Init (&argc, &argv);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
  MPI_Comm_size (MPI_COMM_WORLD, &size);

  if (argc != 6) {
    if (rank == 0) printf("Should be five arguments! Exiting...\n");
    MPI_Finalize ();
    exit(2);
  }

  k      = atof(argv[1]);
  ampl   = atof(argv[2]);
  omd    = atof(argv[3]);
  maxits = atoi(argv[4]);
  dt     = atof(argv[5]);

  if (N % size != 0) {
    if (rank == 0) printf("Numprocs must divide N evenly!\n");
    MPI_Finalize ();
    exit (3);
  }

  /* Number of points to simulate in this subdomain */
  n = N/size;
  m = M;

  /* Allocate arrays */
  x = (double **) calloc (n+2, sizeof(double *));
  x[0] = (double *) calloc ((n+2)*(m+2), sizeof(double));
  for (i=1; i<n+2; i++)
    x[i] = x[i-1] + m + 2;

  v = (double **) calloc (n+2, sizeof(double **));
  v[0] = (double *) calloc ((n+2)*(m+2), sizeof(double));
  for (i=1; i<n+2; i++)
    v[i] = v[i-1] + m + 2;

  /* Set initial values: */
  for (i=0; i<=n+1; i++)
    for (j=0; j<=m+1; j++)
      x[i][j] = 0.0;

  /* Initial (t=0) velocities set here: */
  for (i=1; i<=n; i++)
    for (j=1; j<=m; j++)
      v[i][j] = 0.0;

  /* For half-step method.  This is v[-1/2] = v[0] - (1/2) a[0] dt;
     the first iteration below then produces v[+1/2] as needed. */
  for (i=1; i<=n; i++)
    for (j=1; j<=m; j++)
      v[i][j] += 0.5*k*(4.0*x[i][j] 
			- x[i-1][j] - x[i+1][j]
			- x[i][j-1] - x[i][j+1])*dt;

  /* Determine neighbors (periodic BC possible here!) */
  /* Assumption for now is that ranks increase as we go up the domain;
     p0 is at the bottom */

  dn = rank - 1;
  if (rank < size - 1)
    up = rank + 1;
  else up = -1;

  /* Main loop */
  for (it = 0; it < maxits; it++) {

    /* Write output */
    if (it % OUTSKIP == 0) {
      strcpy (outfile, outbase);
      sprintf (outsuffix, "%04d.%03d", it/OUTSKIP, rank);
      strcat (outfile, outsuffix);

      fp = fopen(outfile, "w");
      for (i=1; i<=n; i++) {
	for (j=1; j<=m; j++)
	  fprintf (fp, "%lf\n", x[i][j]);
	fprintf (fp, "\n"); /* Needed for simple gnuplot display */
      }
      fclose (fp);
    }

    /* Update driving mass(es) */
    if (rank == 0)
      x[0][(m+1)/2] = ampl*sin(omd*(it+1)*dt);

    /* Pass edge values to neighbors.  This should work even if the
       standard send is synchronous, i.e. it does not return until the
       message is received. */

    /* Odds first send, then receive: */
    if (rank % 2 != 0) {
      if (dn >= 0) {
	MPI_Send (&x[1][1],m,MPI_DOUBLE,dn,0,MPI_COMM_WORLD);
	MPI_Recv (&x[0][1],m,MPI_DOUBLE,dn,1,MPI_COMM_WORLD,&stat);
      }
      if (up >= 0) {
	MPI_Send (&x[n][1],m,MPI_DOUBLE,up,1,MPI_COMM_WORLD);
	MPI_Recv (&x[n+1][1],m,MPI_DOUBLE,up,0,MPI_COMM_WORLD,&stat);
      }
    }
    else {
      if (up >= 0) {
	MPI_Recv (&x[n+1][1],m,MPI_DOUBLE,up,0,MPI_COMM_WORLD,&stat);
	MPI_Send (&x[n][1],m,MPI_DOUBLE,up,1,MPI_COMM_WORLD);
      }
      if (dn >= 0) {
	MPI_Recv (&x[0][1],m,MPI_DOUBLE,dn,1,MPI_COMM_WORLD,&stat);
	MPI_Send (&x[1][1],m,MPI_DOUBLE,dn,0,MPI_COMM_WORLD);
      }
    }

    /* Evolution by half step method */
    for (i=1; i<=n; i++)
      for (j=1; j<=m; j++) {
	v[i][j] += - k*(4.0*x[i][j] - x[i-1][j] - x[i+1][j]
			- x[i][j-1] - x[i][j+1])*dt;
      }

    for (i=1; i<=n; i++)
      for (j=1; j<=m; j++)
	x[i][j] += v[i][j]*dt;


  }
  fclose (fp);
  MPI_Finalize ();
  return 0;
}
